import torch.nn as nn
import torch.nn.functional as F


class LabelSmoothingCrossEntropy(nn.Module):
    """标签平滑过的交叉熵loss"""

    def __init__(self, alpha=0.1, reduction='mean', ignore_index=-100):
        super(LabelSmoothingCrossEntropy, self).__init__()
        self.alpha = alpha

        assert reduction in ('mean', 'sum')
        self.reduction = reduction  # 对batch维度的loss的处理策略，可以是mean或者sum

        self.ignore_index = ignore_index  # 忽略target中为ignore_index的位置

    def forward(self, logits, target):
        """计算标签平滑过的交叉熵loss，请参照公式"""

        # 获取类别数K
        K = logits.size()[-1]

        # 对logits计算softmax得到概率，并取log
        log_preds = F.log_softmax(logits, dim=-1)
        # 去掉ignore_index位置的结果
        active_log_preds = log_preds[target != self.ignore_index]

        # 计算非target部分的loss
        if self.reduction == 'sum':
            no_target_loss = -active_log_preds.sum()  # 同时沿着batch和label维度求和
        else:
            no_target_loss = -active_log_preds.sum(dim=-1)  # 沿着label维度求和
            if self.reduction == 'mean':
                no_target_loss = no_target_loss.mean()  # 沿着batch维度平均
        no_target_loss = self.alpha / K * no_target_loss  # 乘上非target部分的yi

        # 计算target部分的loss
        # 借助torch自带的计算交叉熵的函数F.nll_loss，并乘上target部分的yi
        target_loss = (1 - self.alpha) * F.nll_loss(log_preds, target,
                                                    reduction=self.reduction, ignore_index=self.ignore_index)

        return no_target_loss + target_loss



import unittest

import torch

from easy_bert.losses.label_smoothing_loss import LabelSmoothingCrossEntropy


class MyTestCase(unittest.TestCase):
    def test(self):
        print('test~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        lsce = LabelSmoothingCrossEntropy()
        logits = torch.randn(4, 2)  # (batch_size=4, label_size=2)
        target = torch.tensor([0, 1, 1, 0])
        loss = lsce(logits, target)
        print(loss)

    def test_ignore_index(self):
        print('test_ignore_index~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        lsce = LabelSmoothingCrossEntropy(ignore_index=-1)
        logits = torch.randn(6, 2)  # (seq_len=4, label_size=2)
        target = torch.tensor([-1, 0, 1, 1, 0, -1])  # 序列标注一般首尾，即[CLS][SEP]部分用-1填充，计算loss时忽略它们
        loss = lsce(logits, target)
        print(loss)

    def test_reduction(self):
        print('test_reduction~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        lsce = LabelSmoothingCrossEntropy(reduction='sum')
        logits = torch.randn(4, 2)  # (batch_size=4, label_size=2)
        target = torch.tensor([0, 1, 1, 0])
        loss = lsce(logits, target)
        print(loss)


if __name__ == '__main__':
    unittest.main()



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable


class FocalLoss(nn.Module):
    """focal loss"""

    def __init__(self, gamma=0, alpha=None, size_average=True):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        if isinstance(alpha, (float, int, int)):
            self.alpha = torch.Tensor([alpha, 1 - alpha])
        if isinstance(alpha, list):
            self.alpha = torch.Tensor(alpha)
        self.size_average = size_average

    def forward(self, input, target):
        if input.dim() > 2:
            input = input.view(input.size(0), input.size(1), -1)  # N,C,H,W => N,C,H*W
            input = input.transpose(1, 2)  # N,C,H*W => N,H*W,C
            input = input.contiguous().view(-1, input.size(2))  # N,H*W,C => N*H*W,C
        target = target.view(-1, 1)

        # 计算pt
        logpt = F.log_softmax(input)
        logpt = logpt.gather(1, target)
        logpt = logpt.view(-1)
        pt = Variable(logpt.data.exp())

        # 计算log_pt * alpha
        if self.alpha is not None:
            if self.alpha.type() != input.data.type():
                self.alpha = self.alpha.type_as(input.data)
            at = self.alpha.gather(0, target.data.view(-1))  # 根据label，选择对应的label权重
            logpt = logpt * Variable(at)  # log_pt * alpha

        # focal loss 公式
        loss = -1 * (1 - pt) ** self.gamma * logpt

        return loss.mean() if self.size_average else loss.sum()



import os
import re
from functools import partial

from datasets.arrow_dataset import Dataset
from textbrewer import DistillationConfig, TrainingConfig, GeneralDistiller
from torch.utils.data.dataloader import DataLoader
from transformers import AdamW

from easy_bert import logger
from easy_bert.bert4classification.classification_predictor import ClassificationPredictor
from easy_bert.bert4classification.classification_trainer import ClassificationTrainer
from easy_bert.bert4sequence_labeling.sequence_labeling_predictor import SequenceLabelingPredictor
from easy_bert.bert4sequence_labeling.sequence_labeling_trainer import SequenceLabelingTrainer


class TinyBertDistiller(object):
    """
    TinyBert蒸馏器，用12层的bert-base蒸馏4层的TinyBert
    """

    def __init__(self, teacher_pretrained_dir, teacher_model_idr, student_pretrained_dir, student_model_dir,
                 task='classification', enable_parallel=False,
                 hard_label_loss='cross_entropy_loss', focal_loss_gamma=2, focal_loss_alpha=None,
                 temperature=4, hard_label_weight=1,
                 kd_loss_type='ce', kd_loss_weight=1.2, lr=1e-4, ckpt_frequency=1):
        # 初始化目录
        self.teacher_pretrained_dir, self.teacher_model_idr = teacher_pretrained_dir, teacher_model_idr
        self.student_pretrained_dir, self.student_model_dir = student_pretrained_dir, student_model_dir

        # 初始化task
        assert task in ('classification', 'sequence_labeling')
        self.task = task
        self.enable_parallel = enable_parallel

        # 初始化学生hard label的loss_type，即对真实label的loss
        self.hard_label_loss = hard_label_loss
        self.focal_loss_gamma, self.focal_loss_alpha = focal_loss_gamma, focal_loss_alpha

        # 初始化蒸馏配置
        self.temperature = temperature
        self.hard_label_weight = hard_label_weight
        self.kd_loss_type, self.kd_loss_weight = kd_loss_type, kd_loss_weight
        self.lr = lr
        self.ckpt_frequency = ckpt_frequency

        # 分别加载老师model和学生model
        self.teacher_model, self.teacher_predictor = self._load_teacher_model()
        self.student_model, self.student_trainer = self._load_student_model()

    def _load_teacher_model(self):
        """加载老师模型，通过Predictor获取老师模型"""
        logger.info('start to load teacher model ...')
        mapper = {'classification': ClassificationPredictor, 'sequence_labeling': SequenceLabelingPredictor}
        teacher_predictor = mapper[self.task](
            self.teacher_pretrained_dir, self.teacher_model_idr, enable_parallel=self.enable_parallel
        )
        teacher_model = teacher_predictor.model
        teacher_model.forward = partial(teacher_model.forward, return_extra=True)  # 启用return_extra
        logger.info('load teacher model success')
        return teacher_model, teacher_predictor

    def _load_student_model(self):
        """加载学生模型，通过Trainer构建学生模型"""
        logger.info('start to load student model ...')
        mapper = {'classification': ClassificationTrainer, 'sequence_labeling': SequenceLabelingTrainer}
        student_trainer = mapper[self.task](
            self.student_pretrained_dir, self.student_model_dir, learning_rate=self.lr,
            enable_parallel=self.enable_parallel, loss_type=self.hard_label_loss,
            focal_loss_alpha=self.focal_loss_alpha, focal_loss_gamma=self.focal_loss_gamma
        )
        student_trainer.vocab = self.teacher_predictor.vocab  # 学生和老师使用相同的vocab
        student_trainer._build_model()  # 构建模型
        student_trainer.vocab.save_vocab('{}/{}'.format(student_trainer.model_dir, student_trainer.vocab_name))
        student_trainer._save_config()  # 保存训练config
        student_model = student_trainer.model
        student_model.forward = partial(student_model.forward, return_extra=True)  # 启用return_extra
        logger.info('student model load success')
        return student_model, student_trainer

    def distill_train(self, texts, labels, max_len=256, epoch=20, batch_size=32):
        """
        蒸馏训练
        :param texts: list[str] or list[list[str]]
        :param labels: list[str] or list[list[str]]
        :param max_len: 最大长度，超过将被截断
        :param epoch: 迭代轮数，蒸馏一般epoch较大，20~50
        :param batch_size:
        """
        # 封装dataset和dataloader，适配textbrewer接口
        train_input_ids, train_att_mask, train_label_ids = self.student_trainer._transform_batch(
            texts, labels, max_length=max_len
        )
        train_dataset = Dataset.from_dict(
            {'input_ids': train_input_ids, 'attention_mask': train_att_mask, 'labels': train_label_ids}
        )
        train_dataset.set_format(type='torch')  # 设置data type为torch的tensor
        train_dataloader = DataLoader(train_dataset, batch_size=batch_size)

        # 蒸馏配置
        distill_config = DistillationConfig(
            # 设置温度系数temperature, tiny-bert论文作者使用1表现最好，一般大于1比较好
            temperature=self.temperature,
            # 设置ground truth loss权重
            hard_label_weight=self.hard_label_weight,
            # 设置预测层蒸馏loss（即soft label损失）为交叉熵，并稍微放大其权重
            kd_loss_type=self.kd_loss_type, kd_loss_weight=self.kd_loss_weight,
            # 配置中间层蒸馏映射
            intermediate_matches=[
                # 配置hidden蒸馏映射、维度映射
                {'layer_T': 0, 'layer_S': 0, 'feature': 'hidden', 'loss': 'hidden_mse', 'weight': 1,
                 'proj': ['linear', 312, 768]},  # embedding层输出
                {'layer_T': 3, 'layer_S': 1, 'feature': 'hidden', 'loss': 'hidden_mse', 'weight': 1,
                 'proj': ['linear', 312, 768]},
                {'layer_T': 6, 'layer_S': 2, 'feature': 'hidden', 'loss': 'hidden_mse', 'weight': 1,
                 'proj': ['linear', 312, 768]},
                {'layer_T': 9, 'layer_S': 3, 'feature': 'hidden', 'loss': 'hidden_mse', 'weight': 1,
                 'proj': ['linear', 312, 768]},
                {'layer_T': 12, 'layer_S': 4, 'feature': 'hidden', 'loss': 'hidden_mse', 'weight': 1,
                 'proj': ['linear', 312, 768]},
                # 配置attention矩阵蒸馏映射，注意layer序号从0开始
                {"layer_T": 2, "layer_S": 0, "feature": "attention", "loss": "attention_mse", "weight": 1},
                {"layer_T": 5, "layer_S": 1, "feature": "attention", "loss": "attention_mse", "weight": 1},
                {"layer_T": 8, "layer_S": 2, "feature": "attention", "loss": "attention_mse", "weight": 1},
                {"layer_T": 11, "layer_S": 3, "feature": "attention", "loss": "attention_mse", "weight": 1},
            ]
        )

        # 训练配置
        optimizer = AdamW(self.student_model.parameters(), lr=self.lr)  # 使用大一点的lr
        train_config = TrainingConfig(
            output_dir=self.student_model_dir, device=self.student_trainer.device,
            data_parallel=self.enable_parallel, ckpt_frequency=self.ckpt_frequency  # 一个epoch存ckpt_frequency次模型
        )

        # 配置model中logits hiddens attentions losses的获取方法
        def simple_adaptor(batch, model_outputs):
            return {
                'logits': model_outputs[-1]['logits'], 'hidden': model_outputs[-1]['hiddens'],
                'attention': model_outputs[-1]['attentions'], 'losses': model_outputs[1],
            }

        # 蒸馏
        distiller = GeneralDistiller(
            train_config=train_config, distill_config=distill_config,
            model_T=self.teacher_model, model_S=self.student_model,
            adaptor_T=simple_adaptor, adaptor_S=simple_adaptor
        )
        with distiller:
            logger.info('start to knowledge distill ...')
            distiller.train(optimizer, train_dataloader, num_epochs=epoch)
            logger.info('distill finish')

        # 重命名文件名
        newest_model_name = sorted(  # 根据textbrewer模型文件格式，查找最新的模型
            [f for f in os.listdir(self.student_model_dir) if 'pkl' in f],
            key=lambda f: int(re.findall('\\d+', f)[0]), reverse=True
        )[0]
        model_path, new_model_path = \
            '{}/{}'.format(self.student_model_dir, newest_model_name), '{}/bert_model.bin'.format(
                self.student_model_dir)
        logger.info('copy model {} to {}'.format(model_path, new_model_path))
        os.system('cp {} {}'.format(model_path, new_model_path))
        logger.info('now, you can load {} with Predictor'.format(self.student_model_dir))



import torch
import torch.nn as nn

"""
本代码主要来自fastNLP
https://github.com/fastnlp/fastNLP/blob/master/fastNLP/modules/decoder/crf.py
"""


class CRF(nn.Module):
    r"""
    条件随机场。提供forward()以及viterbi_decode()两个方法，分别用于训练与inference。
    """

    def __init__(self, num_tags, include_start_end_trans=False, allowed_transitions=None):
        r"""

        :param int num_tags: 标签的数量
        :param bool include_start_end_trans: 是否考虑各个tag作为开始以及结尾的分数。
        :param List[Tuple[from_tag_id(int), to_tag_id(int)]] allowed_transitions: 内部的Tuple[from_tag_id(int),
                                   to_tag_id(int)]视为允许发生的跃迁，其他没有包含的跃迁认为是禁止跃迁，可以通过
                                   allowed_transitions()函数得到；如果为None，则所有跃迁均为合法
        """
        super(CRF, self).__init__()

        self.include_start_end_trans = include_start_end_trans
        self.num_tags = num_tags

        # the meaning of entry in this matrix is (from_tag_id, to_tag_id) score
        self.trans_m = nn.Parameter(torch.randn(num_tags, num_tags))
        if self.include_start_end_trans:
            self.start_scores = nn.Parameter(torch.randn(num_tags))
            self.end_scores = nn.Parameter(torch.randn(num_tags))

        if allowed_transitions is None:
            constrain = torch.zeros(num_tags + 2, num_tags + 2)
        else:
            constrain = torch.full((num_tags + 2, num_tags + 2), fill_value=-10000.0, dtype=torch.float)
            has_start = False
            has_end = False
            for from_tag_id, to_tag_id in allowed_transitions:
                constrain[from_tag_id, to_tag_id] = 0
                if from_tag_id == num_tags:
                    has_start = True
                if to_tag_id == num_tags + 1:
                    has_end = True
            if not has_start:
                constrain[num_tags, :].fill_(0)
            if not has_end:
                constrain[:, num_tags + 1].fill_(0)
        self._constrain = nn.Parameter(constrain, requires_grad=False)

        nn.init.xavier_normal_(self.trans_m)

    def _normalizer_likelihood(self, logits, mask):
        r"""Computes the (batch_size,) denominator term for the log-likelihood, which is the
        sum of the likelihoods across all possible state sequences.
        :param logits:FloatTensor, max_len x batch_size x num_tags
        :param mask:ByteTensor, max_len x batch_size
        :return:FloatTensor, batch_size
        """
        seq_len, batch_size, n_tags = logits.size()
        alpha = logits[0]
        if self.include_start_end_trans:
            alpha = alpha + self.start_scores.view(1, -1)

        flip_mask = mask.eq(False)

        for i in range(1, seq_len):
            emit_score = logits[i].view(batch_size, 1, n_tags)
            trans_score = self.trans_m.view(1, n_tags, n_tags)
            tmp = alpha.view(batch_size, n_tags, 1) + emit_score + trans_score
            alpha = torch.logsumexp(tmp, 1).masked_fill(flip_mask[i].view(batch_size, 1), 0) + \
                    alpha.masked_fill(mask[i].eq(True).view(batch_size, 1), 0)

        if self.include_start_end_trans:
            alpha = alpha + self.end_scores.view(1, -1)

        return torch.logsumexp(alpha, 1)

    def _gold_score(self, logits, tags, mask):
        r"""
        Compute the score for the gold path.
        :param logits: FloatTensor, max_len x batch_size x num_tags
        :param tags: LongTensor, max_len x batch_size
        :param mask: ByteTensor, max_len x batch_size
        :return:FloatTensor, batch_size
        """
        seq_len, batch_size, _ = logits.size()
        batch_idx = torch.arange(batch_size, dtype=torch.long, device=logits.device)
        seq_idx = torch.arange(seq_len, dtype=torch.long, device=logits.device)

        # trans_socre [L-1, B]
        mask = mask.eq(True)
        flip_mask = mask.eq(False)
        trans_score = self.trans_m[tags[:seq_len - 1], tags[1:]].masked_fill(flip_mask[1:, :], 0)
        # emit_score [L, B]
        emit_score = logits[seq_idx.view(-1, 1), batch_idx.view(1, -1), tags].masked_fill(flip_mask, 0)
        # score [L-1, B]
        score = trans_score + emit_score[:seq_len - 1, :]
        score = score.sum(0) + emit_score[-1].masked_fill(flip_mask[-1], 0)
        if self.include_start_end_trans:
            st_scores = self.start_scores.view(1, -1).repeat(batch_size, 1)[batch_idx, tags[0]]
            last_idx = mask.long().sum(0) - 1
            ed_scores = self.end_scores.view(1, -1).repeat(batch_size, 1)[batch_idx, tags[last_idx, batch_idx]]
            score = score + st_scores + ed_scores
        # return [B,]
        return score

    def forward(self, feats, tags, mask):
        r"""
        用于计算CRF的前向loss，返回值为一个batch_size的FloatTensor，可能需要mean()求得loss。
        :param torch.FloatTensor feats: batch_size x max_len x num_tags，特征矩阵。
        :param torch.LongTensor tags: batch_size x max_len，标签矩阵。
        :param torch.ByteTensor mask: batch_size x max_len，为0的位置认为是padding。
        :return: torch.FloatTensor, (batch_size,)
        """
        feats = feats.transpose(0, 1)
        tags = tags.transpose(0, 1).long()
        mask = mask.transpose(0, 1).float()
        all_path_score = self._normalizer_likelihood(feats, mask)
        gold_path_score = self._gold_score(feats, tags, mask)

        return all_path_score - gold_path_score

    def viterbi_decode(self, logits, mask, unpad=False):
        r"""给定一个特征矩阵以及转移分数矩阵，计算出最佳的路径以及对应的分数
        :param torch.FloatTensor logits: batch_size x max_len x num_tags，特征矩阵。
        :param torch.ByteTensor mask: batch_size x max_len, 为0的位置认为是pad；如果为None，则认为没有padding。
        :param bool unpad: 是否将结果删去padding。False, 返回的是batch_size x max_len的tensor; True，返回的是
            List[List[int]], 内部的List[int]为每个sequence的label，已经除去pad部分，即每个List[int]的长度是这
            个sample的有效长度。
        :return: 返回 (paths, scores)。
                    paths: 是解码后的路径, 其值参照unpad参数.
                    scores: torch.FloatTensor, size为(batch_size,), 对应每个最优路径的分数。
        """
        batch_size, max_len, n_tags = logits.size()
        seq_len = mask.long().sum(1)
        logits = logits.transpose(0, 1).data  # L, B, H
        mask = mask.transpose(0, 1).data.eq(True)  # L, B
        flip_mask = mask.eq(False)

        # dp
        vpath = logits.new_zeros((max_len, batch_size, n_tags), dtype=torch.long)
        vscore = logits[0]  # bsz x n_tags
        transitions = self._constrain.data.clone()
        transitions[:n_tags, :n_tags] += self.trans_m.data
        if self.include_start_end_trans:
            transitions[n_tags, :n_tags] += self.start_scores.data
            transitions[:n_tags, n_tags + 1] += self.end_scores.data

        vscore += transitions[n_tags, :n_tags]

        trans_score = transitions[:n_tags, :n_tags].view(1, n_tags, n_tags).data
        end_trans_score = transitions[:n_tags, n_tags + 1].view(1, 1, n_tags).repeat(batch_size, 1, 1)  # bsz, 1, n_tags

        # 针对长度为1的句子
        vscore += transitions[:n_tags, n_tags + 1].view(1, n_tags).repeat(batch_size, 1) \
            .masked_fill(seq_len.ne(1).view(-1, 1), 0)
        for i in range(1, max_len):
            prev_score = vscore.view(batch_size, n_tags, 1)
            cur_score = logits[i].view(batch_size, 1, n_tags) + trans_score
            score = prev_score + cur_score.masked_fill(flip_mask[i].view(batch_size, 1, 1), 0)  # bsz x n_tag x n_tag
            # 需要考虑当前位置是该序列的最后一个
            score += end_trans_score.masked_fill(seq_len.ne(i + 1).view(-1, 1, 1), 0)

            best_score, best_dst = score.max(1)
            vpath[i] = best_dst
            # 由于最终是通过last_tags回溯，需要保持每个位置的vscore情况
            vscore = best_score.masked_fill(flip_mask[i].view(batch_size, 1), 0) + \
                     vscore.masked_fill(mask[i].view(batch_size, 1), 0)

        # backtrace
        batch_idx = torch.arange(batch_size, dtype=torch.long, device=logits.device)
        seq_idx = torch.arange(max_len, dtype=torch.long, device=logits.device)
        lens = (seq_len - 1)
        # idxes [L, B], batched idx from seq_len-1 to 0
        idxes = (lens.view(1, -1) - seq_idx.view(-1, 1)) % max_len

        ans = logits.new_empty((max_len, batch_size), dtype=torch.long)
        ans_score, last_tags = vscore.max(1)
        ans[idxes[0], batch_idx] = last_tags
        for i in range(max_len - 1):
            last_tags = vpath[idxes[i], batch_idx, last_tags]
            ans[idxes[i + 1], batch_idx] = last_tags
        ans = ans.transpose(0, 1)
        if unpad:
            paths = []
            for idx, max_len in enumerate(lens):
                paths.append(ans[idx, :max_len + 1].tolist())
        else:
            paths = ans
        return paths, ans_score



import unittest

from easy_bert.bert4sequence_labeling.sequence_labeling_predictor import SequenceLabelingPredictor
from easy_bert.bert4sequence_labeling.sequence_labeling_trainer import SequenceLabelingTrainer


class MyTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.model_dir = './tests/test_model'
        # 经测试过的中文bert模型
        self.support_models = [
            'albert_chinese_base', 'chinese-bert-wwm', 'chinese-macbert-base', 'bert-base-chinese',
            'chinese-electra-180g-base-discriminator', 'chinese-roberta-wwm-ext', 'TinyBERT_4L_zh',
            'bert-distil-chinese', 'longformer-chinese-base-4096'
        ]

    def test_general_bert(self):
        print('test_general_bert~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        texts = [
            ['你', '好', '呀'],
            ['一', '马', '当', '先', '就', '是', '好'],
        ]
        labels = [
            ['B', 'E', 'S'],
            ['B', 'M', 'M', 'E', 'S', 'S', 'S']
        ]
        for pretrained_model_dir in self.support_models:
            pretrained_model_dir = './models/{}'.format(pretrained_model_dir)
            print('current test model:', pretrained_model_dir)
            trainer = SequenceLabelingTrainer(pretrained_model_dir, self.model_dir, learning_rate=5e-5)
            trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)
            print('model {} test success'.format(pretrained_model_dir))

    def test_enable_parallel(self):
        print('test_enable_parallel~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        pretrained_model_dir = './models/chinese-roberta-wwm-ext'
        texts = [
            ['你', '好', '呀'],
            ['一', '马', '当', '先', '就', '是', '好'],
        ]
        labels = [
            ['B', 'E', 'S'],
            ['B', 'M', 'M', 'E', 'S', 'S', 'S']
        ]

        trainer = SequenceLabelingTrainer(pretrained_model_dir, self.model_dir, learning_rate=5e-5,
                                          enable_parallel=True)
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

    def test_loss_type(self):
        print('test_loss_type~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        pretrained_model_dir = './models/chinese-roberta-wwm-ext'
        texts = [
            ['你', '好', '呀'],
            ['一', '马', '当', '先', '就', '是', '好'],
        ]
        labels = [
            ['B', 'E', 'S'],
            ['B', 'M', 'M', 'E', 'S', 'S', 'S']
        ]

        trainer = SequenceLabelingTrainer(
            pretrained_model_dir, self.model_dir, learning_rate=5e-5,
            loss_type='focal_loss', focal_loss_gamma=2, focal_loss_alpha=[1, 1, 1, 1]
        )
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

        trainer = SequenceLabelingTrainer(
            pretrained_model_dir, self.model_dir, learning_rate=5e-5,
            loss_type='label_smoothing_loss'
        )
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

        trainer = SequenceLabelingTrainer(
            pretrained_model_dir, self.model_dir, learning_rate=5e-5,
            loss_type='cross_entropy_loss'
        )
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

    def test_adversarial(self):
        print('test_adversarial~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        pretrained_model_dir = './models/chinese-roberta-wwm-ext'
        texts = [
            ['你', '好', '呀'],
            ['一', '马', '当', '先', '就', '是', '好'],
        ]
        labels = [
            ['B', 'E', 'S'],
            ['B', 'M', 'M', 'E', 'S', 'S', 'S']
        ]

        trainer = SequenceLabelingTrainer(pretrained_model_dir, self.model_dir, learning_rate=5e-5, adversarial='fgm')
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

        trainer = SequenceLabelingTrainer(pretrained_model_dir, self.model_dir, learning_rate=5e-5, adversarial='pgd')
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

    def test_predictor(self):
        print('test_predictor~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        pretrained_model_dir = './models/chinese-roberta-wwm-ext'
        texts = [
            ['你', '好', '呀'],
            ['一', '马', '当', '先', '就', '是', '好'],
        ]
        labels = [
            ['B', 'E', 'S'],
            ['B', 'M', 'M', 'E', 'S', 'S', 'S']
        ]
        trainer = SequenceLabelingTrainer(
            pretrained_model_dir, self.model_dir, learning_rate=5e-5, enable_parallel=True
        )
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)
        predictor = SequenceLabelingPredictor(pretrained_model_dir, self.model_dir, enable_parallel=True)
        texts = [
            ['你', '好', '呀'],
            ['一', '马', '当', '先', '就', '是', '好'],
        ]
        labels = predictor.predict(texts)
        print(labels)


if __name__ == '__main__':
    unittest.main()



import time
import unittest

from easy_bert.bert4classification.classification_predictor import ClassificationPredictor
from easy_bert.bert4classification.classification_trainer import ClassificationTrainer
from easy_bert.bert4sequence_labeling.sequence_labeling_predictor import SequenceLabelingPredictor
from easy_bert.bert4sequence_labeling.sequence_labeling_trainer import SequenceLabelingTrainer


class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.model_dir = './tests/test_model'
        self.pretrained_model_dir = './models/chinese-roberta-wwm-ext'

    def test_classification(self):
        print('test_classification~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        texts = ['天气真好', '今天运气很差']
        labels = ['正面', '负面']
        trainer = ClassificationTrainer(self.pretrained_model_dir, self.model_dir, learning_rate=5e-5)
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

        # 测试torch模型速度
        predictor = ClassificationPredictor(self.pretrained_model_dir, self.model_dir)
        init_time = time.time()
        for _ in range(10):
            predictor.predict(texts)
        print('torch model cost time: {}s'.format(time.time() - init_time))

        # 测试onnx模型速度
        predictor.transform2onnx()
        predictor.predict(texts)  # 先预热一次（否则结果可能误导你）
        init_time = time.time()
        for _ in range(10):
            predictor.predict(texts)
        print('onnx model cost time: {}s'.format(time.time() - init_time))

    def test_sequence_labeling(self):
        print('test_sequence_labeling~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        texts = [
            ['你', '好', '呀'],
            ['一', '马', '当', '先', '就', '是', '好'],
        ]
        labels = [
            ['B', 'E', 'S'],
            ['B', 'M', 'M', 'E', 'S', 'S', 'S']
        ]
        trainer = SequenceLabelingTrainer(self.pretrained_model_dir, self.model_dir, learning_rate=5e-5,
                                          loss_type='cross_entropy_loss')  # crf层转onnx暂时有点问题
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

        # 测试torch模型速度
        predictor = SequenceLabelingPredictor(self.pretrained_model_dir, self.model_dir)
        init_time = time.time()
        for _ in range(10):
            predictor.predict(texts)
        print('torch model cost time: {}s'.format(time.time() - init_time))

        # 测试onnx模型速度
        predictor.transform2onnx()
        predictor.predict(texts)  # 先预热一次（否则结果可能误导你）
        init_time = time.time()
        for _ in range(10):
            predictor.predict(texts)
        print('onnx model cost time: {}s'.format(time.time() - init_time))


if __name__ == '__main__':
    unittest.main()



import abc


class BaseTrainer(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def train(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def validate(self, *args, **kwargs):
        pass



import unittest

from easy_bert.bert4classification.classification_predictor import ClassificationPredictor
from easy_bert.bert4classification.classification_trainer import ClassificationTrainer
from easy_bert.bert4sequence_labeling.sequence_labeling_predictor import SequenceLabelingPredictor
from easy_bert.bert4sequence_labeling.sequence_labeling_trainer import SequenceLabelingTrainer
from easy_bert.tinybert_distiller import TinyBertDistiller


class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.teacher_pretrained = './models/chinese-roberta-wwm-ext'
        self.teacher_model_dir = './tests/test_model'
        self.student_pretrained = './models/TinyBERT_4L_zh'
        self.student_model_dir = './tests/test_model2'

    def test_general(self):
        print('test_classification~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        texts = [
            '天气真好',
            '今天运气很差',
        ]
        labels = ['正面', '负面']

        # 训练老师模型
        trainer = ClassificationTrainer(self.teacher_pretrained, self.teacher_model_dir)
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

        # 蒸馏学生
        distiller = TinyBertDistiller(
            self.teacher_pretrained, self.teacher_model_dir, self.student_pretrained, self.student_model_dir,
            task='classification'
        )
        distiller.distill_train(texts, labels, max_len=20, epoch=20, batch_size=2)

        # 加载fine-tune蒸馏过的模型
        predictor = ClassificationPredictor(self.student_pretrained, self.student_model_dir)
        print(predictor.predict(texts))

    def test_sequence_labeling(self):
        print('test_sequence_labeling~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        texts = [
            ['你', '好', '呀'],
            ['一', '马', '当', '先', '就', '是', '好'],
        ]
        labels = [
            ['B', 'E', 'S'],
            ['B', 'M', 'M', 'E', 'S', 'S', 'S']
        ]

        # 训练老师模型
        trainer = SequenceLabelingTrainer(self.teacher_pretrained, self.teacher_model_dir, loss_type='crf_loss')
        trainer.train(texts, labels, validate_texts=texts, validate_labels=labels, batch_size=2, epoch=20)

        # 蒸馏学生
        distiller = TinyBertDistiller(
            self.teacher_pretrained, self.teacher_model_dir, self.student_pretrained, self.student_model_dir,
            task='sequence_labeling', hard_label_loss='crf_loss'
        )
        distiller.distill_train(texts, labels, max_len=20, epoch=20, batch_size=2)

        # 加载fine-tune蒸馏过的模型
        predictor = SequenceLabelingPredictor(self.student_pretrained, self.student_model_dir)
        print(predictor.predict(texts))


if __name__ == '__main__':
    unittest.main()



import json
import math
import os
import random

import numpy as np
import torch
import torch.cuda
from sklearn.utils import shuffle
from torch.nn import DataParallel
from transformers import AdamW

from easy_bert import logger
from easy_bert.adversarial import FGM, PGD
from easy_bert.base.base_trainer import BaseTrainer
from easy_bert.bert4classification.classification_model import ClassificationModel
from easy_bert.vocab import Vocab


class ClassificationTrainer(BaseTrainer):
    def __init__(self, pretrained_model_dir, model_dir, learning_rate=5e-5, ckpt_name='bert_model.bin',
                 vocab_name='vocab.json', enable_parallel=False, adversarial=None, dropout_rate=0.5,
                 loss_type='cross_entropy_loss', focal_loss_gamma=2, focal_loss_alpha=None, random_seed=0):
        self.pretrained_model_dir = pretrained_model_dir
        self.model_dir = model_dir
        self.ckpt_name = ckpt_name
        self.vocab_name = vocab_name
        self.enable_parallel = enable_parallel

        # 设置随机种子
        self.random_seed = random_seed
        self._set_random_seed(random_seed)

        self.adversarial = adversarial
        assert adversarial in (None, 'fgm', 'pgd')

        self.loss_type = loss_type
        self.focal_loss_gamma = focal_loss_gamma
        self.focal_loss_alpha = focal_loss_alpha

        # 自动获取当前设备
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self.dropout_rate = dropout_rate
        self.learning_rate = learning_rate
        self.batch_size = None
        self.epoch = None

        self.vocab = Vocab()

    def _set_random_seed(self, seed):
        """针对torch torch.cuda numpy random分别设定随机种子"""
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        np.random.seed(seed)
        random.seed(seed)

    def _build_model(self):
        """构建bert模型"""
        # 实例化ClassificationModel模型
        self.model = ClassificationModel(
            self.pretrained_model_dir, self.vocab.label_size, dropout_rate=self.dropout_rate,
            loss_type=self.loss_type, focal_loss_alpha=self.focal_loss_alpha, focal_loss_gamma=self.focal_loss_gamma
        )

        # 设置AdamW优化器
        no_decay = ["bias", "LayerNorm.weight"]  # bias和LayerNorm不使用正则化
        # 参数分两组，分为decay or no_decay
        optimizer_grouped_parameters = [
            {'params': [p for n, p in self.model.named_parameters() if not any(nd in n for nd in no_decay)],
             'weight_decay': 0.01},
            {'params': [p for n, p in self.model.named_parameters() if any(nd in n for nd in no_decay)],
             'weight_decay': 0.0},
        ]
        self.optimizer = AdamW(optimizer_grouped_parameters, lr=self.learning_rate)

        # 使用bert的vocab更新Vocab对象
        self.vocab.set_vocab2id(self.model.get_bert_tokenizer().vocab)
        self.vocab.set_id2vocab({_id: char for char, _id in self.vocab.vocab2id.items()})
        self.vocab.set_unk_vocab_id(self.vocab.vocab2id['[UNK]'])
        self.vocab.set_pad_vocab_id(self.vocab.vocab2id['[PAD]'])

        # 启用并行，使用DataParallel封装model
        if self.enable_parallel:
            self.model = DataParallel(self.model, device_ids=list(range(torch.cuda.device_count())))

        # 将模型拷贝到当前设备
        self.model.to(self.device)

    def _save_config(self):
        """保存训练参数配置为json文件"""
        config = {
            'vocab_size': self.vocab.vocab_size,
            'label_size': self.vocab.label_size,
            'learning_rate': self.learning_rate,
            'batch_size': self.batch_size,
            'dropout_rate': self.dropout_rate,
            'epoch': self.epoch,
            'ckpt_name': self.ckpt_name,
            'vocab_name': self.vocab_name,
            'enable_parallel': self.enable_parallel,
            'adversarial': self.adversarial,
            'loss_type': self.loss_type,
            'focal_loss_gamma': self.focal_loss_gamma,
            'focal_loss_alpha': self.focal_loss_alpha,
            'pretrained_model': os.path.basename(self.pretrained_model_dir),
        }
        with open('{}/train_config.json'.format(self.model_dir), 'w') as f:
            f.write(json.dumps(config, indent=4))

    def _transform_batch(self, batch_texts, batch_labels, max_length=512):
        """将batch的文本及labels转换为bert的输入tensor形式"""
        batch_input_ids, batch_att_mask, batch_label_ids = [], [], []
        for text, label in zip(batch_texts, batch_labels):  # 迭代每条样本
            assert isinstance(text, str)
            # 根据是否并行，获取bert_tokenizer
            bert_tokenizer = self.model.bert_tokenizer if not self.enable_parallel else self.model.module.bert_tokenizer
            encoded_dict = bert_tokenizer.encode_plus(text, max_length=max_length, padding='max_length',
                                                      return_tensors='pt', truncation=True)
            batch_input_ids.append(encoded_dict['input_ids'])
            batch_att_mask.append(encoded_dict['attention_mask'])
            batch_label_ids.append(self.vocab.tag2id[label])
        batch_input_ids, batch_att_mask, batch_label_ids = \
            torch.cat(batch_input_ids), torch.cat(batch_att_mask), torch.LongTensor(batch_label_ids)

        # 将数据拷贝到当前设备
        batch_input_ids, batch_att_mask, batch_label_ids = \
            batch_input_ids.to(self.device), batch_att_mask.to(self.device), batch_label_ids.to(self.device)

        return batch_input_ids, batch_att_mask, batch_label_ids

    def train(self, train_texts, train_labels, validate_texts, validate_labels,
              batch_size=30, epoch=10, max_loss_num=10, warning_max_len=True):
        """训练
        Args:
            train_texts: list[str] 训练集样本
            train_labels: list[str] 训练集标签
            validate_texts: list[str] 验证集样本
            validate_labels: list[str] 验证集标签
            batch_size: int
            epoch: int
            max_loss_num: int 连续max_loss_num个validate loss平均，根据其保存模型
            warning_max_len: bool 是否警告max_len超过512（普通bert不能处理超过512，除了一些变体如longformer）
        """
        self.batch_size = batch_size
        self.epoch = epoch

        # 构建词库表（只构建label id映射，词库bert已有）、构建模型
        self.vocab.build_vocab(labels=train_labels, build_texts=False, with_build_in_tag_id=False)
        self._build_model()
        # 保存词库、保存训练config
        self.vocab.save_vocab('{}/{}'.format(self.model_dir, self.vocab_name))
        self._save_config()

        logger.info('train samples: {}, validate samples: {}'.format(len(train_texts), len(validate_texts)))

        best_loss = float("inf")  # 当前最小的loss
        loss_buff = []  # 保存最近max_loss_num的个valid loss

        step = 0

        # 根据对抗配置，实例化不同的对抗实例
        if self.adversarial:
            logger.info('enable adversarial training')
            adv = FGM(self.model) if self.adversarial == 'fgm' else PGD(self.model)

        for epoch in range(epoch):
            for batch_idx in range(math.ceil(len(train_texts) / batch_size)):
                text_batch = train_texts[batch_size * batch_idx: batch_size * (batch_idx + 1)]
                labels_batch = train_labels[batch_size * batch_idx: batch_size * (batch_idx + 1)]

                step += 1
                self.model.train()  # 设置为train模式
                self.model.zero_grad()  # 清空梯度

                # 训练
                batch_max_len = max([len(text) for text in text_batch]) + 2  # 长度得加上[CLS]和[SEP]
                if warning_max_len and batch_max_len > 512:
                    logger.warning(
                        'current batch max_len is {}, > 512, which may not be processed by bert!'.format(batch_max_len)
                    )
                batch_input_ids, batch_att_mask, batch_label_ids = self._transform_batch(text_batch,
                                                                                         labels_batch,
                                                                                         max_length=batch_max_len)
                best_paths, loss = self.model(batch_input_ids, batch_att_mask, labels=batch_label_ids)
                if self.enable_parallel:  # 如果启用并行，需将多张卡返回的sub-batch loss平均
                    loss = loss.mean()
                loss.backward()

                # 对抗训练
                if self.adversarial:
                    adv.train(batch_input_ids, batch_att_mask, labels=batch_label_ids)

                # 更新参数
                self.optimizer.step()

                # 计算train acc、valid acc
                train_acc = self._get_acc_one_step(best_paths, batch_label_ids)
                valid_acc, valid_loss = self.validate(validate_texts, validate_labels, batch_size, warning_max_len)

                # 计算连续max_loss_num个平均valid_loss
                loss_buff.append(valid_loss)
                if len(loss_buff) > max_loss_num:
                    loss_buff = loss_buff[-max_loss_num:]
                avg_loss = sum(loss_buff) / len(loss_buff) if len(loss_buff) == max_loss_num else None

                logger.info(
                    'epoch %d, step %d, train loss %.4f, train acc %.4f, valid loss %.4f valid acc %.4f, '
                    'last %d avg valid loss %s' % (
                        epoch, step, loss, train_acc, valid_loss, valid_acc, max_loss_num,
                        '%.4f' % avg_loss if avg_loss else avg_loss
                    )
                )

                # 如果avg_loss在降，保存模型
                if avg_loss and avg_loss < best_loss:
                    best_loss = avg_loss
                    # 根据是否启用并行，获得state_dict
                    state_dict = self.model.state_dict() if not self.enable_parallel else self.model.module.state_dict()
                    torch.save(state_dict, '{}/{}'.format(self.model_dir, self.ckpt_name))
                    logger.info("model saved")

        logger.info("finished")

    def validate(self, validate_texts, validate_labels, sample_size=100, warning_max_len=True):
        """验证
        Args:
            validate_texts: list[str]
            validate_labels: list[str]
            sample_size: int 采样大小(使用全量验证集较慢，这里每次随机采样sample_size个样本做验证)
            warning_max_len: bool 是否警告max_len超过512（普通bert不能处理超过512，除了一些变体如longformer）
        Returns:
            float 验证集上acc, loss
        """
        # 设置为evaluate模式
        self.model.eval()

        # 随机采样sample_size个样本
        batch_texts, batch_labels = [
            return_val[:sample_size] for return_val in shuffle(validate_texts, validate_labels)
        ]

        # 计算valid acc, valid loss
        batch_max_len = max([len(text) for text in batch_texts]) + 2
        if warning_max_len and batch_max_len > 512:
            logger.warning(
                'current batch max_len is {}, > 512, which may not be processed by bert!'.format(batch_max_len)
            )
        with torch.no_grad():  # eval时不计算梯度
            batch_input_ids, batch_att_mask, batch_label_ids = self._transform_batch(batch_texts,
                                                                                     batch_labels,
                                                                                     max_length=batch_max_len)
            best_paths, loss = self.model(batch_input_ids, batch_att_mask, labels=batch_label_ids)
            if self.enable_parallel:
                loss = loss.mean()
            acc = self._get_acc_one_step(best_paths, batch_label_ids)
            return acc, loss

    def _get_acc_one_step(self, labels_predict_batch, labels_batch):
        """计算一个batch的acc, correct_label_num / total_label_num"""
        acc = (labels_predict_batch == labels_batch).sum().item() / labels_batch.shape[0]
        return float(acc)



import logging
import sys

logger = logging.getLogger('easy-bert')
logger.setLevel(logging.DEBUG)

handler = logging.StreamHandler(stream=sys.stdout)
handler.setLevel(logging.DEBUG)

formatter = logging.Formatter(
    '%(asctime)s - %(levelname)s - %(filename)s:%(lineno)s - %(funcName)s() - %(message)s')

handler.setFormatter(formatter)

logger.addHandler(handler)



import json
import math

import torch
import torch.onnx
from onnxruntime import InferenceSession
from torch.nn import DataParallel

from easy_bert.base.base_predictor import BasePredictor
from easy_bert.bert4sequence_labeling.sequence_labeling_model import SequenceLabelingModel
from easy_bert.vocab import Vocab


class SequenceLabelingPredictor(BasePredictor):
    def __init__(self, pretrained_model_dir, model_dir, vocab_name='vocab.json',
                 enable_parallel=False):
        self.pretrained_model_dir = pretrained_model_dir
        self.model_dir = model_dir
        self.enable_parallel = enable_parallel

        # 自动获取当前设备
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self.vocab = Vocab()

        # 加载config、vocab、model
        self._load_config()
        self.vocab.load_vocab('{}/{}'.format(model_dir, vocab_name))
        self._load_model()

        # onnx相关配置
        self.is_onnx_model = False  # 当前模型是否为onnx模型
        self.onnx_model = None  # onnx_model初始化为None

    def _load_config(self):
        """json加载训练config"""
        with open('{}/train_config.json'.format(self.model_dir), 'r') as f:
            self._config = json.loads(f.read())

    def _load_model(self):
        """加载模型"""
        # 根据config初始化SequenceLabelingModel
        # 注意，部分参数推理时并不需要，这里传入debug时可读性更好
        self.model = SequenceLabelingModel(
            self.pretrained_model_dir, self._config['label_size'], self._config['dropout_rate'],
            loss_type=self._config['loss_type'], focal_loss_alpha=self._config['focal_loss_alpha'],
            focal_loss_gamma=self._config['focal_loss_gamma']
        )

        # 加载模型文件里的参数
        self.model.load_state_dict(
            torch.load('{}/{}'.format(self.model_dir, self._config['ckpt_name']), map_location=self.device)
        )

        # 设置为evaluate模式
        self.model.eval()

        # 如果启用并行，使用DataParallel封装model
        if self.enable_parallel:
            self.model = DataParallel(self.model, device_ids=list(range(torch.cuda.device_count())))

        # 将模型拷贝到当前设备
        self.model.to(self.device)

        # 手动设置vocab中unk、pad标签id
        self.vocab.set_unk_vocab_id(self.vocab.vocab2id['[UNK]'])
        self.vocab.set_pad_vocab_id(self.vocab.vocab2id['[PAD]'])

    def predict(self, texts, batch_size=64, max_len=512):
        """
        Args:
            texts: list[list[str]] 预测样本
            batch_size: int
            max_len: int 最大序列长度
        Returns:
            list[list[str]] 标签序列
        """
        batch_labels = []

        for batch_idx in range(math.ceil(len(texts) / batch_size)):
            text_batch = texts[batch_size * batch_idx: batch_size * (batch_idx + 1)]

            # 当前batch最大长度
            batch_max_len = min(max([len(text) for text in text_batch]) + 2, max_len)

            # 将texts转化为tensor
            batch_input_ids, batch_att_mask = [], []
            for text in text_batch:
                assert isinstance(text, list)
                text = ' '.join(text)  # 确保输入encode_plus函数为文本
                bert_tokenizer = self.get_bert_tokenizer()
                encoded_dict = bert_tokenizer.encode_plus(text, max_length=batch_max_len,
                                                          padding='max_length',
                                                          return_tensors='pt', truncation=True)
                batch_input_ids.append(encoded_dict['input_ids'])
                batch_att_mask.append(encoded_dict['attention_mask'])
            batch_input_ids, batch_att_mask = torch.cat(batch_input_ids), torch.cat(batch_att_mask)

            # 将数据拷贝到当前设备
            batch_input_ids, batch_att_mask = batch_input_ids.to(self.device), batch_att_mask.to(self.device)

            # 推理，并将结果解析为原始label
            with torch.no_grad():
                if self.is_onnx_model:  # onnx model推理
                    best_paths = self.onnx_model.run(
                        ['output'],  # 设置输出names
                        # feed输入，并转化为numpy数组
                        {'input1': batch_input_ids.cpu().numpy(), 'input2': batch_att_mask.cpu().numpy()}
                    )[0]
                    best_paths = torch.IntTensor(best_paths).to(self.device)  # numpy数组转为tensor，保持接口一致
                else:  # torch模型推理
                    best_paths = self.model(batch_input_ids, batch_att_mask)
                for best_path, att_mask in zip(best_paths, batch_att_mask):
                    active_labels = best_path[att_mask == 1][1:-1]  # 截掉pad、[CLS]、[SEP]部分
                    labels = [self.vocab.id2tag[label_id.item()] for label_id in active_labels]
                    batch_labels.append(labels)

        return batch_labels

    def get_bert_tokenizer(self):
        """根据是否并行，获取bert_tokenizer"""
        bert_tokenizer = self.model.bert_tokenizer if not self.enable_parallel else self.model.module.bert_tokenizer
        return bert_tokenizer

    def transform2onnx(self):
        """将模型转换为onnx模型"""
        assert not self.is_onnx_model, 'error, curl model is already onnx model!'

        # 定义伪输入，让onnx做一遍推理，构建静态计算图
        dummy_inputs = torch.LongTensor([[i for i in range(200)]])
        dummy_att_masks = torch.LongTensor([[1 for _ in range(200)]])
        dummy_inputs, dummy_att_masks = dummy_inputs.to(self.device), dummy_att_masks.to(self.device)

        # 将模型导出为onnx标准
        torch.onnx.export(
            self.model, (dummy_inputs, dummy_att_masks),
            '{}/model.onnx'.format(self.model_dir),
            # 设置model的输入输出，参考SequenceLabelingModel.forward函数签名
            input_names=['input1', 'input2'], output_names=['output'],  # 两个输入，一个输出
            # 设置batch、seq_len维度可变
            dynamic_axes={
                'input1': {0: 'batch', 1: 'seq'}, 'input2': {0: 'batch', 1: 'seq'}, 'output': {0: 'batch', 1: 'seq'}},
            opset_version=10,
        )

        # 加载onnx model
        self._load_onnx_model()

    def _load_onnx_model(self):
        """加载onnx model"""
        # 通过InferenceSession，加载onnx模型
        self.onnx_model = InferenceSession('{}/model.onnx'.format(self.model_dir))
        self.onnx_model.get_modelmeta()

        # 更新is_cur_onnx_model状态
        self.is_onnx_model = True



import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss
from transformers import AlbertModel
from transformers import BertTokenizer, BertModel
from transformers import DistilBertTokenizer, DistilBertModel
from transformers import ElectraTokenizer, ElectraModel
from transformers import LongformerModel, LongformerTokenizer
from transformers.activations import get_activation

from easy_bert.losses.focal_loss import FocalLoss
from easy_bert.losses.label_smoothing_loss import LabelSmoothingCrossEntropy


class ClassificationModel(nn.Module):

    def __init__(self, bert_base_model_dir, label_size, dropout_rate=0.5,
                 loss_type='cross_entropy_loss', focal_loss_gamma=2, focal_loss_alpha=None):
        super(ClassificationModel, self).__init__()
        self.label_size = label_size

        assert loss_type in ('cross_entropy_loss', 'focal_loss', 'label_smoothing_loss')
        if focal_loss_alpha:  # 确保focal_loss_alpha合法，必须是一个label的概率分布
            assert isinstance(focal_loss_alpha, list) and len(focal_loss_alpha) == label_size
        self.loss_type = loss_type
        self.focal_loss_gamma = focal_loss_gamma
        self.focal_loss_alpha = focal_loss_alpha

        # 自动获取当前设备
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        # 根据预训练文件名，自动检测bert的各种变体，并加载
        if 'albert' in bert_base_model_dir.lower():
            self.bert_tokenizer = BertTokenizer.from_pretrained(bert_base_model_dir)
            self.bert_model = AlbertModel.from_pretrained(bert_base_model_dir)
        elif 'electra' in bert_base_model_dir.lower():
            self.bert_tokenizer = ElectraTokenizer.from_pretrained(bert_base_model_dir)
            self.bert_model = ElectraModel.from_pretrained(bert_base_model_dir)
        elif 'longformer' in bert_base_model_dir.lower():
            self.bert_tokenizer = BertTokenizer.from_pretrained(bert_base_model_dir)
            # longformer-chinese-base-4096模型参数prefix为bert而非标准的longformer，这是个坑
            LongformerModel.base_model_prefix = 'bert'
            self.bert_model = LongformerModel.from_pretrained(bert_base_model_dir)
        elif 'distil' in bert_base_model_dir.lower():
            self.bert_tokenizer = DistilBertTokenizer.from_pretrained(bert_base_model_dir)
            self.bert_model = DistilBertModel.from_pretrained(bert_base_model_dir)
        else:
            self.bert_tokenizer = BertTokenizer.from_pretrained(bert_base_model_dir)
            self.bert_model = BertModel.from_pretrained(bert_base_model_dir)

        self.dropout = nn.Dropout(dropout_rate)
        # 定义linear层，进行 hidden_size->hidden_size 的映射，可以使用其作为bert pooler
        # 原始bert pooler是一个使用tanh激活的全连接层
        self.linear = nn.Linear(self.bert_model.config.hidden_size, self.bert_model.config.hidden_size)
        # 全连接分类层
        self.cls_layer = nn.Linear(self.bert_model.config.hidden_size, label_size)

    def forward(self, input_ids, attention_mask, token_type_ids=None, position_ids=None,
                head_mask=None, inputs_embeds=None, labels=None, return_extra=False):

        # Longformer的[cls]位置设置全局Attention，值为2
        if isinstance(self.bert_model, LongformerModel):
            attention_mask[:, 0] = 2
            bert_out = self.bert_model(
                input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids,
                position_ids=position_ids, inputs_embeds=None, return_dict=False,
                output_hidden_states=True,  # longformer直接output_attentions会报错，这里暂时将其attentions置为None
            ) + (None,)
        elif isinstance(self.bert_model, DistilBertModel):
            # distilbert不支持token_type_ids、position_ids参数，不传入
            bert_out = self.bert_model(
                input_ids=input_ids, attention_mask=attention_mask, inputs_embeds=None, return_dict=False,
                output_hidden_states=True, output_attentions=True
            )
        else:
            bert_out = self.bert_model(
                input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids,
                position_ids=position_ids, inputs_embeds=None, return_dict=False,
                output_hidden_states=True, output_attentions=True
            )
        bert_out, (hidden_states, attentions) = bert_out[:-2], bert_out[-2:]

        if isinstance(self.bert_model, ElectraModel):
            # Electra使用gelu激活的bert pooler
            last_hidden_state, = bert_out
            x = last_hidden_state[:, 0, :]  # 取[CLS]位置的输出
            x = self.dropout(x)
            x = self.linear(x)
            x = get_activation("gelu")(x)
            x = self.dropout(x)
            pooled_output = x
        elif isinstance(self.bert_model, DistilBertModel):
            # distilbert使用relu激活的bert pooler
            last_hidden_state, = bert_out
            x = last_hidden_state[:, 0, :]
            x = self.linear(x)
            x = get_activation("relu")(x)
            x = self.dropout(x)
            pooled_output = x
        else:
            last_hidden_state, pooled_output = bert_out[:2]
        cls_outs = self.dropout(pooled_output)
        logits = self.cls_layer(cls_outs)

        best_labels = torch.argmax(logits, dim=-1).to(self.device)

        # 将logits hiddens attentions装进extra，蒸馏时可能需要使用
        extra = {'hiddens': hidden_states, 'logits': logits, 'attentions': attentions}

        if labels is not None:
            # 根据不同的loss_type，选择不同的loss计算
            if self.loss_type == 'cross_entropy_loss':
                loss = CrossEntropyLoss()(logits, labels)
            elif self.loss_type == 'focal_loss':
                loss = FocalLoss(gamma=self.focal_loss_gamma, alpha=self.focal_loss_alpha)(logits, labels)
            else:
                loss = LabelSmoothingCrossEntropy(alpha=0.1)(logits, labels)
            return (best_labels, loss) if not return_extra else (best_labels, loss, extra)

        return best_labels if not return_extra else (best_labels, extra)

    def get_bert_tokenizer(self):
        return self.bert_tokenizer



from setuptools import setup, find_packages

setup(
    name='easy-zh-bert',
    version='0.4.0',
    author='waking95',
    author_email="wang0624@foxmail.com",
    license='MIT',
    description="easy-bert是一个中文NLP工具，提供诸多bert变体调用和调参方法，极速上手；清晰的设计和代码注释，也很适合学习",
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/waking95/easy-bert",
    project_urls={
        "Bug Tracker": "https://github.com/waking95/easy-bert/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    packages=find_packages(),
    install_requires=[
        'torch == 1.4.0',
        'transformers == 3.1.0',
        'scikit-learn == 0.24.0',
        'numpy == 1.17.0',
        'datasets == 1.16.0',
        'textbrewer == 0.2.0',
        # 'onnxruntime==1.4.0',
        'onnxruntime-gpu == 1.4.0',
    ],
)



import torch

"""
对抗训练

在embedding输出上加扰动，即在embedding层权重上加扰动
embedding层即一个特殊的全连接层，输入onehot查询向量 seq_len * vocab_size，权重 vocab_size * embedding_size
"""


class FGM(object):
    """
    Fast Gradient Method（FGM）
    """

    def __init__(self, model, epsilon=1., emb_name='word_embeddings'):
        self.model = model
        self.epsilon = epsilon
        self.emb_name = emb_name

        self.backup = {}

    def attack(self):
        """对抗，计算embedding层扰动r，并相加"""
        for name, param in self.model.named_parameters():
            if param.requires_grad and self.emb_name in name:  # 找到embedding层参数
                self.backup[name] = param.data.clone()  # 备份参数到backup
                norm = torch.norm(param.grad)  # 计算2范数
                if norm != 0 and not torch.isnan(norm):
                    r_adv = self.epsilon * param.grad / norm  # 计算fgm扰动r
                    param.data.add_(r_adv)  # x + r

    def restore(self):
        """恢复embedding层参数"""
        for name, param in self.model.named_parameters():
            if param.requires_grad and self.emb_name in name:
                assert name in self.backup
                param.data = self.backup[name]  # 从backup中还原原始embedding层参数
        self.backup = {}  # 清空backup

    def train(self, *args, **kwargs):
        """对抗训练"""
        self.attack()  # 在embedding上增加扰动
        _, loss_adv = self.model(*args, **kwargs)  # 使用扰动后的embedding计算新loss
        loss_adv.backward()  # 反向传播，在正常梯度上，累加扰动后的梯度
        self.restore()  # 恢复embedding层参数


class PGD(object):
    """
    Projected Gradient Descent（PGD）
    """

    def __init__(self, model, epsilon=1., alpha=0.3, emb_name='word_embeddings', k=3):
        self.model = model  # 对抗的model
        self.epsilon = epsilon  # 最大扰动半径
        self.alpha = alpha  # 步长
        self.emb_name = emb_name
        self.k = k  # 对抗阶数

        self.emb_backup = {}  # 备份embedding参数
        self.grad_backup = {}  # 备份梯度

    def attack(self, is_first_attack=False):
        """对抗，计算embedding层扰动r，并相加"""
        for name, param in self.model.named_parameters():
            if param.requires_grad and self.emb_name in name:  # 找到embedding层参数
                if is_first_attack:  # 第一次对抗时，备份embedding参数
                    self.emb_backup[name] = param.data.clone()
                norm = torch.norm(param.grad)  # 计算2范数
                if norm != 0 and not torch.isnan(norm):
                    r_adv = self.alpha * param.grad / norm  # 计算扰动r
                    param.data.add_(r_adv)  # x + r
                    param.data = self.project(name, param.data, self.epsilon)  # 扰动半径超过epsilon，投影回来

    def restore(self):
        """恢复embedding层参数"""
        for name, param in self.model.named_parameters():
            if param.requires_grad and self.emb_name in name:
                assert name in self.emb_backup
                param.data = self.emb_backup[name]
        self.emb_backup = {}

    def project(self, param_name, param_data, epsilon):
        """投影，确保扰动r不超过扰动半径"""
        r = param_data - self.emb_backup[param_name]  # 获得当前扰动半径r
        if torch.norm(r) > epsilon:  # 如果扰动半径过大，投影（二范数归一）
            r = epsilon * r / torch.norm(r)
        return self.emb_backup[param_name] + r  # 返回新的 x + r

    def backup_grad(self):
        """备份对抗前所有参数梯度"""
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                self.grad_backup[name] = param.grad.clone()

    def restore_grad(self):
        """恢复对抗前所有参数梯度"""
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                param.grad = self.grad_backup[name]

    def train(self, *args, **kwargs):
        """对抗训练"""
        # 备份梯度
        self.backup_grad()

        # 进行k阶对抗
        # 前k-1次model call主要计算扰动r
        # 最后1次model call使用扰动后的embedding计算对抗loss，累加梯度
        for k_i in range(self.k):
            # 对抗，对embedding添加扰动，（第1次对抗时备份embedding参数）
            self.attack(is_first_attack=(k_i == 0))
            if k_i != self.k - 1:
                self.model.zero_grad()  # 前几次对抗时，清空梯度，便于计算k_i时刻新梯度
            else:
                self.restore_grad()  # 最后一次对抗时，恢复对抗前所有参数梯度
            _, loss_adv = self.model(*args, **kwargs)  # 使用扰动后的embedding计算新loss
            loss_adv.backward()  # 反向传播，计算新梯度（最后一次时，累加扰动的梯度）

        self.restore()  # 恢复embedding层参数



import json

from easy_bert import logger


class Vocab(object):
    """
    词汇表

    基于texts构建vocab2id id2vocab，可以是char or word
    基于labels构建tag2id id2tag
    使用bert时，无需重新构建vocab2id id2vocab，直接set_vocab2id set_id2vocab即可
    """

    def __init__(self, vocab2id=None, id2vocab=None, tag2id=None, id2tag=None,
                 unk_vocab_id=0, pad_vocab_id=1, pad_tag_id=0):
        self.vocab2id = vocab2id
        self.id2vocab = id2vocab
        self.tag2id = tag2id
        self.id2tag = id2tag
        self.vocab_size = 0 if not vocab2id else len(vocab2id)
        self.label_size = 0 if not tag2id else len(tag2id)
        self.unk_vocab_id = unk_vocab_id
        self.pad_vocab_id = pad_vocab_id
        self.pad_tag_id = pad_tag_id

    def build_vocab(self, texts=None, labels=None, build_texts=True, build_labels=True, with_build_in_tag_id=True):
        """构建词汇表"""
        logger.info('start to build vocab ...')

        if build_texts:
            assert texts, 'Please make sure texts is not None!'
            self.vocab2id, self.id2vocab = {'<UKN>': self.unk_vocab_id, '<PAD>': self.pad_vocab_id}, \
                                           {self.unk_vocab_id: '<UKN>', self.pad_vocab_id: '<PAD>'}
            vocab_cnt = 2
            for text in texts:
                for seg in text:
                    if seg in self.vocab2id:
                        continue
                    self.vocab2id[seg] = vocab_cnt
                    self.id2vocab[vocab_cnt] = seg
                    vocab_cnt += 1
            self.vocab_size = len(self.vocab2id)

        if build_labels:
            assert labels, 'Please make sure labels is not None!'
            self.tag2id, self.id2tag = {'<PAD>': self.pad_tag_id}, {self.pad_tag_id: '<PAD>'}
            tag_cnt = 1
            if not with_build_in_tag_id:  # label不预置PAD_ID
                self.tag2id, self.id2tag = {}, {}
                tag_cnt = 0
            for label in labels:
                if isinstance(label, list):
                    for each_label in label:
                        if each_label in self.tag2id:
                            continue
                        self.tag2id[each_label] = tag_cnt
                        self.id2tag[tag_cnt] = each_label
                        tag_cnt += 1
                else:  # label为str类型
                    assert isinstance(label, str)
                    self.tag2id[label] = tag_cnt
                    self.id2tag[tag_cnt] = label
                    tag_cnt += 1
            self.label_size = len(self.tag2id)

        logger.info('build vocab finish, vocab_size: {}, label_size: {}'.format(self.vocab_size, self.label_size))

    def save_vocab(self, vocab_file):
        """保存词汇表"""
        result = {
            'vocab2id': self.vocab2id,
            'id2vocab': self.id2vocab,
            'tag2id': self.tag2id,
            'id2tag': self.id2tag
        }
        with open(vocab_file, 'w', encoding='utf-8') as f:
            f.write(json.dumps(result, ensure_ascii=False, indent=4))
        logger.info('save vocab to {}'.format(vocab_file))

    def load_vocab(self, vocab_file):
        """加载词汇表"""
        with open(vocab_file, 'r', encoding='utf-8') as f:
            result = json.loads(f.read())
        self.vocab2id = result['vocab2id']
        self.id2vocab = {int(k): v for k, v in result['id2vocab'].items()}  # 将id转为int
        self.tag2id = result['tag2id']
        self.id2tag = {int(k): v for k, v in result['id2tag'].items()}
        self.vocab_size = len(self.vocab2id)
        self.label_size = len(self.tag2id)

    def set_vocab2id(self, vocab2id):
        self.vocab2id = vocab2id
        self.vocab_size = len(self.vocab2id)
        return self

    def set_id2vocab(self, id2vocab):
        self.id2vocab = id2vocab
        return self

    def set_tag2id(self, tag2id):
        self.tag2id = tag2id
        self.label_size = len(self.tag2id)
        return self

    def set_id2tag(self, id2tag):
        self.id2tag = id2tag
        return self

    def set_unk_vocab_id(self, unk_vocab_id):
        self.unk_vocab_id = unk_vocab_id

    def set_pad_vocab_id(self, pad_vocab_id):
        self.pad_vocab_id = pad_vocab_id

    def set_pad_tag_id(self, pad_tag_id):
        self.pad_tag_id = pad_tag_id



import unittest

import torch

from easy_bert.losses.crf_layer import CRF

torch.manual_seed(1)
torch.cuda.manual_seed_all(1)


class MyTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.crf = CRF(4)
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.crf.to(self.device)

    def test_decode(self):
        print('test_decode~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        logits = torch.randn((2, 3, 4), requires_grad=False).to(self.device)  # (batch_size, seq_len, num_label)
        mask = torch.IntTensor([[1, 1, 1], [1, 1, 0]]).to(self.device)  # 第二个样本pad了一个位置
        labels, scores = self.crf.viterbi_decode(logits, mask)
        print(labels)

    def test_loss(self):
        print('test_loss~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        logits = torch.randn((2, 3, 4), requires_grad=True).to(self.device)  # (batch_size, seq_len, num_label)
        mask = torch.IntTensor([[1, 1, 1], [1, 1, 0]]).to(self.device)  # 第二个样本pad了一个位置
        labels = torch.LongTensor([[0, 2, 3], [1, 0, 1]]).to(self.device)
        loss = self.crf.forward(logits, labels, mask)
        print(loss)


if __name__ == '__main__':
    unittest.main()



import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss
from transformers import AlbertModel
from transformers import BertTokenizer, BertModel
from transformers import DistilBertTokenizer, DistilBertModel
from transformers import ElectraTokenizer, ElectraModel
from transformers import LongformerModel

from easy_bert.losses.crf_layer import CRF
from easy_bert.losses.focal_loss import FocalLoss
from easy_bert.losses.label_smoothing_loss import LabelSmoothingCrossEntropy


class SequenceLabelingModel(nn.Module):

    def __init__(self, bert_base_model_dir, label_size, drop_out_rate=0.5,
                 loss_type='crf_loss', focal_loss_gamma=2, focal_loss_alpha=None):
        super(SequenceLabelingModel, self).__init__()
        self.label_size = label_size

        assert loss_type in ('crf_loss', 'cross_entropy_loss', 'focal_loss', 'label_smoothing_loss')
        if focal_loss_alpha:  # 确保focal_loss_alpha合法，必须是一个label的概率分布
            assert isinstance(focal_loss_alpha, list) and len(focal_loss_alpha) == label_size
        self.loss_type = loss_type
        self.focal_loss_gamma = focal_loss_gamma
        self.focal_loss_alpha = focal_loss_alpha

        # 自动获取当前设备
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        # 根据预训练文件名，自动检测bert的各种变体，并加载
        if 'albert' in bert_base_model_dir.lower():
            self.bert_tokenizer = BertTokenizer.from_pretrained(bert_base_model_dir)
            self.bert_model = AlbertModel.from_pretrained(bert_base_model_dir)
        elif 'electra' in bert_base_model_dir.lower():
            self.bert_tokenizer = ElectraTokenizer.from_pretrained(bert_base_model_dir)
            self.bert_model = ElectraModel.from_pretrained(bert_base_model_dir)
        elif 'longformer' in bert_base_model_dir.lower():
            self.bert_tokenizer = BertTokenizer.from_pretrained(bert_base_model_dir)
            # longformer-chinese-base-4096模型参数prefix为bert而非标准的longformer，这是个坑
            LongformerModel.base_model_prefix = 'bert'
            self.bert_model = LongformerModel.from_pretrained(bert_base_model_dir)
        elif 'distil' in bert_base_model_dir.lower():
            self.bert_tokenizer = DistilBertTokenizer.from_pretrained(bert_base_model_dir)
            self.bert_model = DistilBertModel.from_pretrained(bert_base_model_dir)
        else:
            self.bert_tokenizer = BertTokenizer.from_pretrained(bert_base_model_dir)
            self.bert_model = BertModel.from_pretrained(bert_base_model_dir)

        self.dropout = nn.Dropout(drop_out_rate)
        # 定义linear层，将hidden_size映射到label_size层
        self.linear = nn.Linear(self.bert_model.config.hidden_size, label_size)
        # 定义crf层
        self.crf = CRF(label_size)

    def forward(self, input_ids, attention_mask, token_type_ids=None, position_ids=None,
                head_mask=None, inputs_embeds=None, labels=None, return_extra=False):

        if isinstance(self.bert_model, LongformerModel):
            bert_out = self.bert_model(
                input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids,
                position_ids=position_ids, inputs_embeds=None, return_dict=False,
                output_hidden_states=True,  # longformer直接output_attentions会报错，这里暂时将其attentions置为None
            ) + (None,)
        elif isinstance(self.bert_model, DistilBertModel):
            # distilbert不支持token_type_ids、position_ids参数，不传入
            bert_out = self.bert_model(
                input_ids=input_ids, attention_mask=attention_mask, inputs_embeds=None, return_dict=False,
                output_hidden_states=True, output_attentions=True
            )
        else:
            bert_out = self.bert_model(
                input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids,
                position_ids=position_ids, inputs_embeds=None, return_dict=False,
                output_hidden_states=True, output_attentions=True
            )
        bert_out, (hidden_states, attentions) = bert_out[:-2], bert_out[-2:]

        last_hidden_state = bert_out[0]
        seq_outs = self.dropout(last_hidden_state)
        logits = self.linear(seq_outs)

        # 根据loss_type，选择使用维特比解码或直接argmax
        if self.loss_type == 'crf_loss':
            best_paths, scores = self.crf.viterbi_decode(logits, attention_mask)
        else:
            best_paths = torch.argmax(logits, dim=-1)

        # 将logits hiddens attentions装进extra，蒸馏时可能需要使用
        extra = {'hiddens': hidden_states, 'logits': logits, 'attentions': attentions}

        if labels is not None:
            # 根据不同的loss_type，选择不同的loss计算
            active_loss = attention_mask.view(-1) == 1  # 通过attention_mask忽略pad
            active_logits, active_labels = logits.view(-1, self.label_size)[active_loss], labels.view(-1)[active_loss]
            if self.loss_type == 'crf_loss':
                # 计算loss时，忽略[CLS]、[SEP]以及PAD部分
                lengths, new_att_mask = attention_mask.sum(axis=1), attention_mask.clone()
                for i, length in enumerate(lengths):
                    new_att_mask[i, length - 1] = 0  # [SEP]部分置为0
                loss = self.crf.forward(logits[:, 1:, :], labels[:, 1:], new_att_mask[:, 1:]).mean()
            elif self.loss_type == 'cross_entropy_loss':
                loss = CrossEntropyLoss(ignore_index=-1)(active_logits, active_labels)  # 忽略label=-1位置
            elif self.loss_type == 'focal_loss':
                active_loss = active_labels != -1  # 忽略label=-1位置，即[CLS]和[SEP]
                active_logits, active_labels = active_logits[active_loss], active_labels[active_loss]
                loss = FocalLoss(gamma=self.focal_loss_gamma, alpha=self.focal_loss_alpha)(active_logits, active_labels)
            else:
                loss = LabelSmoothingCrossEntropy(alpha=0.1, ignore_index=-1)(active_logits, active_labels)
            return (best_paths, loss) if not return_extra else (best_paths, loss, extra)

        return best_paths if not return_extra else (best_paths, extra)

    def get_bert_tokenizer(self):
        return self.bert_tokenizer
